/*
  format.h - String Format Library for ARDUINO

  Copyright (c) 2024 Sasapea's Lab. All right reserved.

  This program is free software: you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 3 of
  the License, or at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program. If not, see <https://www.gnu.org/licenses/>.
*/
#pragma once

#include <stdarg.h>
#include <stdlib.h>
#include <stdint.h>

#if defined(ARDUINO_ARCH_AVR)
  #if !defined(FORMAT_SUPPORT_FLOAT_PRECISON)
    #define FORMAT_SUPPORT_FLOAT_PRECISON 0
  #endif
  typedef int ssize_t;
#endif

#if !defined(FORMAT_SUPPORT_SSIZE_T)
  #define FORMAT_SUPPORT_SSIZE_T 1
#endif

#if !defined(FORMAT_SUPPORT_FLOAT)
  #define FORMAT_SUPPORT_FLOAT 1
#endif

#if !defined(FORMAT_SUPPORT_FLOAT_PRECISON)
  #define FORMAT_SUPPORT_FLOAT_PRECISON 1
#endif

class Format
{
  public:

    int printf(const char *format, ...)
    {
      va_list ap;
      va_start(ap, format);
      int nbytes = vprintf(format, ap);
      va_end(ap);
      return nbytes;
    }

    int vprintf(const char *format, va_list ap)
    {
      const char *p = format;
      char c;
      _count = 0;
      do
      {
        c = *p++;
        if (c == '%')
        {
          format_t option = {};
          option.alignchar = (*p == '-' ? *p++ : 0);
          option.signchar  = (*p == '+' ? *p++ : 0);
          option.fillchar  = (*p == '0' ? *p++ : ' ');
          option.precision = -1;
          if ((*p >= '0') && (*p <= '9'))
            option.width = strtoul(p, (char **)&p, 10);
          else if (*p == '*')
          {
            ++p;
            option.width = va_arg(ap, int);
          }
          if (*p == '.')
          {
            ++p;
            if ((*p >= '0') && (*p <= '9'))
              option.precision = strtoul(p, (char **)&p, 10);
            else if (*p == '*')
            {
              ++p;
              option.precision = va_arg(ap, int);
            }
          }
          switch (*p)
          {
            case 'L':
            case 'l':
              do
              {
                ++p;
                if (option.type < TYPE_LONG_LONG_INT)
                  ++option.type;
              } while ((*p == 'l') || (*p == 'L'));
              break;
            case 'z':
              ++p;
              option.type = TYPE_SIZE_T;
              break;
            case 'h':
              do
              {
                ++p;
                if (option.type < TYPE_LONG_LONG_INT)
                  ++option.type;
              } while (*p == 'h');
              option.type = TYPE_I16 - (option.type - 1);
              break;
            case 'I':
              switch (strtoul(++p, (char **)&p, 10))
              {
                case 8:
                  option.type = TYPE_I8;
                  break;
                case 16:
                  option.type = TYPE_I16;
                  break;
                default:
                case 32:
                  option.type = TYPE_I32;
                  break;
                case 64:
                  option.type = TYPE_I64;
                  break;
              }
              break;
            case 'j':
              ++p;
              option.type = TYPE_IMAX;
              break;
          }
          switch (*p)
          {
            case 'd':
            case 'i':
            case 'u':
            case 'X':
            case 'x':
            case 'o':
            case 'b':
              switch (option.type)
              {
                default: option.type = TYPE_INT; /* fall through */
                case TYPE_INT          : option.value.i     = va_arg(ap, int); break;
                case TYPE_LONG_INT     : option.value.li    = va_arg(ap, long int); break;
                case TYPE_LONG_LONG_INT: option.value.lli   = va_arg(ap, long long int); break;
                case TYPE_SIZE_T       : option.value.isize = va_arg(ap, size_t); break;
                case TYPE_I8           : option.value.i8    = va_arg(ap, int); break;
                case TYPE_I16          : option.value.i16   = va_arg(ap, int); break;
                case TYPE_I32          : option.value.i32   = va_arg(ap, int32_t); break;
                case TYPE_I64          : option.value.i64   = va_arg(ap, int64_t); break;
                case TYPE_IMAX         : option.value.imax  = va_arg(ap, intmax_t); break;
              }
              break;
            case 'f':
            case 'F':
              switch (option.type)
              {
#if FORMAT_SUPPORT_FLOAT
                default: option.type = TYPE_INT; /* fall through */
                case TYPE_INT     : option.value.f  = va_arg(ap, double); break;
                case TYPE_LONG_INT: option.value.lf = va_arg(ap, long double); break;
#else
                default: option.type = TYPE_INT; /* fall through */
                case TYPE_INT     : va_arg(ap, double); break;
                case TYPE_LONG_INT: va_arg(ap, long double); break;
#endif
              }
              option.type += TYPE_DOUBLE;
              break;
            default:
              option.type = TYPE_INT;
              break;
          }
          switch (*p++)
          {
            case 'd': /* %d - integer */
            case 'i': /* %i - integer */
            case 'f': /* %f - floating */
            case 'F': /* %F - floating */
              switch (option.type)
              {
                default:  /* fall through */
                case TYPE_INT:
                  if ((int)option.value.i < 0)
                  {
                    option.value.i = 0 - option.value.i;
                    option.signchar = '-';
                  }
                  break;
                case TYPE_LONG_INT:
                  if ((long int)option.value.li < 0)
                  {
                    option.value.li = 0 - option.value.li;
                    option.signchar = '-';
                  }
                  break;
                case TYPE_LONG_LONG_INT:
                  if ((long long int)option.value.lli < 0)
                  {
                    option.value.lli = 0 - option.value.lli;
                    option.signchar = '-';
                  }
                  break;
#if FORMAT_SUPPORT_SSIZE_T
                case TYPE_SIZE_T:
                  if ((ssize_t)option.value.isize < 0)
                  {
                    option.value.isize = 0 - option.value.isize;
                    option.signchar = '-';
                  }
                  break;
#endif
                case TYPE_I8:
                  if ((int8_t)option.value.i8 < 0)
                  {
                    option.value.i8 = 0 - option.value.i8;
                    option.signchar = '-';
                  }
                  break;
                case TYPE_I16:
                  if ((int16_t)option.value.i16 < 0)
                  {
                    option.value.i16 = 0 - option.value.i16;
                    option.signchar = '-';
                  }
                  break;
                case TYPE_I32:
                  if ((int32_t)option.value.i32 < 0)
                  {
                    option.value.i32 = 0 - option.value.i32;
                    option.signchar = '-';
                  }
                  break;
                case TYPE_I64:
                  if ((int64_t)option.value.i64 < 0)
                  {
                    option.value.i64 = 0 - option.value.i64;
                    option.signchar = '-';
                  }
                  break;
                case TYPE_IMAX:
                  if ((intmax_t)option.value.imax < 0)
                  {
                    option.value.imax = 0 - option.value.imax;
                    option.signchar = '-';
                  }
                  break;
#if FORMAT_SUPPORT_FLOAT
                case TYPE_DOUBLE:
                  if (option.value.f < 0)
                  {
                    option.value.f = 0 - option.value.f;
                    option.signchar = '-';
                  }
                  break;
                case TYPE_LONG_DOUBLE:
                  if (option.value.lf < 0)
                  {
                    option.value.lf = 0 - option.value.lf;
                    option.signchar = '-';
                  }
                  break;
#endif
              }
              option.base = 10;
              nformat(option);
              break;
            case 'u': /* %u - unsigned integer */
              option.base = 10;
              nformat(option);
              break;
            case 'X': /* %X - hex */
              option.hexchar = 'A';
              option.base = 16;
              nformat(option);
              break;
            case 'x': /* %x - hex */
              option.hexchar = 'a';
              option.base = 16;
              nformat(option);
              break;
            case 'o': /* %o - octal */
              option.base = 8;
              nformat(option);
              break;
            case 'b': /* %b - binary */
              option.base = 2;
              nformat(option);
              break;
            case 'c': /* %c - character */
              putch_(va_arg(ap, int));
              break;
            case 's': /* %s - string */
              option.fillchar = ' ';
              putstr(option, va_arg(ap, char *));
              break;
            case '%': /* %% - '%' character */
              putch_('%');
              break;
            default:
              putch_('?');
              break;
          }
        }
        else if (c)
          putch_(c);
      }
      while (c);
      return _count;
    }

  protected:

    size_t _count;

    virtual int putch(char c) = 0;

  private:

    enum
    {
      TYPE_INT,
      TYPE_LONG_INT,
      TYPE_LONG_LONG_INT,
      TYPE_SIZE_T,
      TYPE_I8,
      TYPE_I16,
      TYPE_I32,
      TYPE_I64,
      TYPE_IMAX,
      TYPE_DOUBLE,
      TYPE_LONG_DOUBLE
    };

    typedef struct
    {
      char alignchar;
      char signchar;
      char fillchar;
      char hexchar;
      size_t width;
      int    precision;
      unsigned char type;
      unsigned char base;
      union
      {
        unsigned int i;
        unsigned long int li;
        unsigned long long int lli;
        size_t isize;
        uint8_t i8;
        uint16_t i16;
        uint32_t i32;
        uint64_t i64;
        uintmax_t imax;
#if FORMAT_SUPPORT_FLOAT
        double f;
        long double lf;
#endif
      } value;
    } format_t;

    void putch_(char c)
    {
      _count += putch(c);
    }

    void putstr(format_t& option, const char *str)
    {
      if (!str)
        str = "(null)";
      size_t i, len = strlen(str);
      if ((option.precision >= 0) && (len > (size_t)option.precision))
        len = option.precision;
      if (option.alignchar == '-')
      {
        for (i = 0; i < len; ++i)
          putch_(*str++);
        for (i = len; i < option.width; ++i)
          putch_(option.fillchar);
      }
      else
      {
        for (i = len; i < option.width; ++i)
          putch_(option.fillchar);
        for (i = 0; i < len; ++i)
          putch_(*str++);
      }
    }

    void nformat(format_t& option)
    {
      char buf[128], c, z, *p = buf + sizeof(buf);
      unsigned char n;
#if FORMAT_SUPPORT_FLOAT_PRECISON == 0
      uint32_t d;
      static constexpr int DEFAULT_PRECISION = 3;
      static constexpr typeof(d) POW10[] = {
        1,
        10,
        100,
        1000,
        10000,
        100000,
        1000000,
        10000000 /* 7 */
      };
#else
      uint64_t d;
      static constexpr int DEFAULT_PRECISION = 6;
      static constexpr typeof(d) POW10[] = {
        1,
        10,
        100,
        1000,
        10000,
        100000,
        1000000,
        10000000,
        100000000,
        1000000000,
        10000000000,
        100000000000,
        1000000000000,
        10000000000000,
        100000000000000,
        1000000000000000,
        10000000000000000 /* 16 */
      };
#endif
      static constexpr unsigned char POW10_LIMIT = sizeof(POW10) / sizeof(POW10[0]) - 1;
      *--p = 0;
      switch (option.type)
      {
        case TYPE_DOUBLE:
          z = option.precision <= 0;
          n = option.precision < 0 ? DEFAULT_PRECISION : (option.precision < POW10_LIMIT ? option.precision : POW10_LIMIT);
#if FORMAT_SUPPORT_FLOAT
          d = POW10[n];
          d = (typeof(d))((option.value.f += 0.5 / d) * d) % d;
#else
          d = 0;
#endif
          while (d)
          {
            c = d % option.base;
            d /= option.base;
            if (c || !z)
            {
              *--p = '0' + c;
              z = 0;
            }
            --n;
          }
          if (!z)
          {
            while (n--)
              *--p = '0';
            *--p = '.';
          }
          option.precision = -1;
          option.type = TYPE_I64;
#if FORMAT_SUPPORT_FLOAT
          option.value.i64 = option.value.f;
#else
          option.value.i64 = 0;
#endif
          break;
        case TYPE_LONG_DOUBLE:
          z = option.precision <= 0;
          n = option.precision < 0 ? DEFAULT_PRECISION : (option.precision < POW10_LIMIT ? option.precision : POW10_LIMIT);
#if FORMAT_SUPPORT_FLOAT
          d = POW10[n];
          d = (typeof(d))((option.value.lf += 0.5 / d) * d) % d;
#else
          d = 0;
#endif
          while (d)
          {
            c = d % option.base;
            d /= option.base;
            if (c || !z)
            {
              *--p = '0' + c;
              z = 0;
            }
            --n;
          }
          if (!z)
          {
            while (n--)
              *--p = '0';
            *--p = '.';
          }
          option.precision = -1;
          option.type = TYPE_IMAX;
#if FORMAT_SUPPORT_FLOAT
          option.value.imax = option.value.lf;
#else
          option.value.imax = 0;
#endif
          break;
      }
      do
      {
        switch (option.type)
        {
          default:  /* fall through */
          case TYPE_INT:
            c = option.value.i % option.base;
            z = (option.value.i /= option.base) != 0;
            break;
          case TYPE_LONG_INT:
            c = option.value.li % option.base;
            z = (option.value.li /= option.base) != 0;
            break;
          case TYPE_LONG_LONG_INT:
            c = option.value.lli % option.base;
            z = (option.value.lli /= option.base) != 0;
            break;
          case TYPE_SIZE_T:
            c = option.value.isize % option.base;
            z = (option.value.isize /= option.base) != 0;
            break;
          case TYPE_I8:
            c = option.value.i8 % option.base;
            z = (option.value.i8 /= option.base) != 0;
            break;
          case TYPE_I16:
            c = option.value.i16 % option.base;
            z = (option.value.i16 /= option.base) != 0;
            break;
          case TYPE_I32:
            c = option.value.i32 % option.base;
            z = (option.value.i32 /= option.base) != 0;
            break;
          case TYPE_I64:
            c = option.value.i64 % option.base;
            z = (option.value.i64 /= option.base) != 0;
            break;
          case TYPE_IMAX:
            c = option.value.imax % option.base;
            z = (option.value.imax /= option.base) != 0;
            break;
        }
        if (c < 10)
          *--p = '0' + c;
        else
          *--p = option.hexchar + (c - 10);
      }
      while (z);
      if (option.signchar)
        *--p = option.signchar;
      putstr(option, p);
    }
};

class StringFormat : public Format
{
  public:

    int snprintf(char *buf, size_t n, const char *format, ...)
    {
      va_list ap;
      va_start(ap, format);
      int nbytes = vsnprintf(buf, n, format, ap);
      va_end(ap);
      return nbytes;
    }

    int vsnprintf(char *buf, size_t n, const char *format, va_list ap)
    {
      _buffer = buf;
      _length = n;
      int nbytes = vprintf(format, ap);
      buf[nbytes] = 0;
      return nbytes;
    }

  protected:

    int putch(char c) override
    {
      if (_count + 1 < _length)
      {
        *_buffer++ = c;
        return 1;
      }
      return 0;
    }

  private:

    char  *_buffer;
    size_t _length;
};

#define DECLARE_PRINTF(instance) PrintFormat<typeof(instance), instance>

template<typename T, T &UART>
class PrintFormat : public Format
{
  protected:

    int putch(char c) override
    {
      if (c == '\n')
        UART.println();
      else
        UART.print(c);
      return 1;
    }
};
